"""Lists of physical variables made available by StagPy.

They are organized by kind of variables (field, profiles, and time series).
`EXTRA` lists group variables that are not directly output by StagYY and need to
be computed from other variables.
"""

from __future__ import annotations

import typing
from operator import attrgetter
from types import MappingProxyType

from . import processing
from .datatypes import Varf, Varr, Vart

if typing.TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Callable

    from .datatypes import Field, Rprof, Tseries
    from .dimensions import Scales
    from .stagyydata import StagyyData
    from .step import Step


FIELD: Mapping[str, Varf] = MappingProxyType(
    {
        "T": Varf("Temperature", "K"),
        "v1": Varf("Velocity (x)", "m/s"),
        "v2": Varf("Velocity (y)", "m/s"),
        "v3": Varf("Velocity (z)", "m/s"),
        "p": Varf("Pressure", "Pa"),
        "eta": Varf("Viscosity", "Pa.s"),
        "rho": Varf("Density", "kg/m3"),
        "rho4rhs": Varf("Density term in RHS", "kg/m3"),
        "trarho": Varf("Density from tracer mass", "kg/m3"),
        "sII": Varf("Second invariant of stress tensor", "Pa"),
        "sx1": Varf("Principal stress eigenvector (x)", "Pa"),
        "sx2": Varf("Principal stress eigenvector (y)", "Pa"),
        "sx3": Varf("Principal stress eigenvector (z)", "Pa"),
        "s1val": Varf("Principal stress eigenvalue", "Pa"),
        "edot": Varf("Strain rate", "1/s"),
        "Tcond": Varf("Thermal conductivity", "W/(m.K)"),
        "c": Varf("Composition", "1"),
        "cFe": Varf("FeO content", "1"),
        "hpe": Varf("HPE content", "1"),
        "wtr": Varf("Water concentration", "1"),
        "age": Varf("Age", "s"),
        "contID": Varf("ID of continents", "1"),
        "rs1": Varf("Momentum residue (x)", "1"),
        "rs2": Varf("Momentum residue (y)", "1"),
        "rs3": Varf("Momentum residue (z)", "1"),
        "rsc": Varf("Continuity residue", "1"),
        "basalt": Varf("Basalt fraction", "1"),
        "harzburgite": Varf("Harzburgite fraction", "1"),
        "impactor": Varf("Impactor fraction", "1"),
        "prim": Varf("Primordial layer", "1"),
        "meltfrac": Varf("Melt fraction", "1"),
        "meltcompo": Varf("Melt composition", "1"),
        "meltrate": Varf("Melting rate", "1"),
        "meltvel": Varf("Melt velocity", "1"),
        "nmelt": Varf("N melt", "1"),
        "fSiO2": Varf("fSiO2", "1"),
        "fMgO": Varf("fMgO", "1"),
        "fFeO": Varf("fFeO", "1"),
        "fXO": Varf("fXO", "1"),
        "fFeR": Varf("fFeR", "1"),
    }
)
"""Scalar fields output by StagYY."""

FIELD_EXTRA: Mapping[str, Callable[[Step], Field]] = MappingProxyType(
    {
        "stream": processing.stream_function,
    }
)
"""Scalar fields that StagPy can compute."""

FIELD_FILES: Mapping[str, list[str]] = MappingProxyType(
    {
        "t": ["T"],
        "vp": ["v1", "v2", "v3", "p"],
        "c": ["c"],
        "eta": ["eta"],
        "rho": ["rho"],
        "hpe": ["hpe"],
        "wtr": ["wtr"],
        "age": ["age"],
        "nrc": ["contID"],
        "rs": ["rs1", "rs2", "rs3", "rsc"],
        "str": ["sII"],
        "sx": ["sx1", "sx2", "sx3", "s1val"],
        "ed": ["edot"],
        "k": ["Tcond"],
        "bs": ["basalt"],
        "hz": ["harzburgite"],
        "imp": ["impactor"],
        "prm": ["prim"],
        "f": ["meltfrac"],
        "fc": ["meltcompo"],
        "df": ["meltrate"],
        "vm": ["meltvel"],
        "nm": ["nmelt"],
        "fSiO2": ["fSiO2"],
        "fMgO": ["fMgO"],
        "fFeO": ["fFeO"],
        "fXO": ["fXO"],
        "fFeR": ["fFeR"],
    }
)

FIELD_FILES_H5: Mapping[str, list[str]] = MappingProxyType(
    {
        "Temperature": ["T"],
        "Velocity": ["v1", "v2", "v3"],
        "Dynamic_Pressure": ["p"],
        # Depending on the version of StagYY and whether total pressure is used,
        # the dynamic pressure might be written in this file instead.  It will be
        # used as fallback by step.Fields._get_raw_data.
        "Pressure": ["p"],
        "Composition": ["c"],
        "IronContent": ["cFe"],
        "HPE": ["hpe"],
        "Viscosity": ["eta"],
        "Density": ["rho"],
        "Density4rhs": ["rho4rhs"],
        "TraBasedDensity": ["trarho"],
        "water": ["wtr"],
        "Age": ["age"],
        "ContinentNumber": ["contID"],
        "ResidualMomentum": ["rs1", "rs2", "rs3"],
        "ResidualContinuity": ["rsc"],
        "Stress": ["sII"],
        "PrincipalStressAxis": ["sx1", "sx2", "sx3"],
        "StrainRate": ["edot"],
        "ThermalConductivity": ["Tcond"],
        "Basalt": ["basalt"],
        "Harzburgite": ["harzburgite"],
        "Impactor": ["impactor"],
        "Primordial": ["prim"],
        "MeltFrac": ["meltfrac"],
        "MeltComposition": ["meltcompo"],
        "MeltingRate": ["meltrate"],
        "MeltVelocity": ["meltvel"],
        "nmelt": ["nmelt"],
        "fSiO2": ["fSiO2"],
        "fMgO": ["fMgO"],
        "fFeO": ["fFeO"],
        "fXO": ["fXO"],
        "fFeR": ["fFeR"],
    }
)

SFIELD: Mapping[str, Varf] = MappingProxyType(
    {
        "topo_top": Varf("Topography at top", "m"),
        "topo_bot": Varf("Topography at bottom", "m"),
        "geoid_top": Varf("Geoid at top", "m"),
        "geoid_bot": Varf("Geoid at bottom", "m"),
        "topo_g_top": Varf("Topography for geoid at top", "m"),
        "topo_g_bot": Varf("Topography for geoid at bottom", "m"),
        "ftop": Varf("Heat flux at top", "W/m2"),
        "fbot": Varf("Heat flux at bottom", "W/m2"),
        "fstop": Varf("Heat flux from spectrum at top", "W/m2"),
        "fsbot": Varf("Heat flux from spectrum at bottom", "W/m2"),
        "crust": Varf("Crustal thickness", "m"),
    }
)
"""Surface scalar fields output by StagYY."""

SFIELD_FILES: Mapping[str, list[str]] = MappingProxyType(
    {
        "cs": ["topo_bot", "topo_top"],
        "g": ["geoid_bot", "geoid_top"],
        "csg": ["topo_g_bot", "topo_g_top"],
        "hf": ["fbot", "ftop"],
        "hfs": ["fsbot", "fstop"],
        "cr": ["crust"],
    }
)

SFIELD_FILES_H5: Mapping[str, list[str]] = MappingProxyType(
    {
        "BottomTopography": ["topo_bot"],
        "SurfaceTopography": ["topo_top"],
        "BottomGeoid": ["geoid_bot"],
        "TopGeoid": ["geoid_top"],
        "BottomCSGeoid": ["topo_g_bot"],
        "TopCSGeoid": ["topo_g_top"],
        "BottomHeatFlux": ["fbot"],
        "TopHeatFlux": ["ftop"],
        "BottomHFSpectrum": ["fsbot"],
        "TopHFSpectrum": ["fstop"],
        "CrustThickness": ["crust"],
    }
)


RPROF: Mapping[str, Varr] = MappingProxyType(
    {
        "r": Varr("Radial coordinate", "Radius", "m"),
        "Tmean": Varr("Temperature", "Temperature", "K"),
        "Tmin": Varr("Min temperature", "Temperature", "K"),
        "Tmax": Varr("Max temperature", "Temperature", "K"),
        "vrms": Varr("rms velocity", "Velocity", "m/s"),
        "vmin": Varr("Min velocity", "Velocity", "m/s"),
        "vmax": Varr("Max velocity", "Velocity", "m/s"),
        "vzabs": Varr("Radial velocity", "Velocity", "m/s"),
        "vzmin": Varr("Min radial velocity", "Velocity", "m/s"),
        "vzmax": Varr("Max radial velocity", "Velocity", "m/s"),
        "vhrms": Varr("Horizontal velocity", "Velocity", "m/s"),
        "vhmin": Varr("Min horiz velocity", "Velocity", "m/s"),
        "vhmax": Varr("Max horiz velocity", "Velocity", "m/s"),
        "etalog": Varr("Viscosity", "Viscosity", "Pa"),
        "etamin": Varr("Min viscosity", "Viscosity", "Pa"),
        "etamax": Varr("Max viscosity", "Viscosity", "Pa"),
        "elog": Varr("Strain rate", "Strain rate", "1/s"),
        "emin": Varr("Min strain rate", "Strain rate", "1/s"),
        "emax": Varr("Max strain rate", "Strain rate", "1/s"),
        "slog": Varr("Stress", "Stress", "Pa"),
        "smin": Varr("Min stress", "Stress", "Pa"),
        "smax": Varr("Max stress", "Stress", "Pa"),
        "whrms": Varr("Horizontal vorticity", "Vorticity", "1/s"),
        "whmin": Varr("Min horiz vorticity", "Vorticity", "1/s"),
        "whmax": Varr("Max horiz vorticity", "Vorticity", "1/s"),
        "wzrms": Varr("Radial vorticity", "Vorticity", "1/s"),
        "wzmin": Varr("Min radial vorticity", "Vorticity", "1/s"),
        "wzmax": Varr("Max radial vorticity", "Vorticity", "1/s"),
        "drms": Varr("Divergence", "Divergence", "1/s"),
        "dmin": Varr("Min divergence", "Divergence", "1/s"),
        "dmax": Varr("Max divergence", "Divergence", "1/s"),
        "enadv": Varr("Advection", "Heat flux", "W/m2"),
        "endiff": Varr("Diffusion", "Heat flux", "W/m2"),
        "enradh": Varr("Radiogenic heating", "Heat flux", "W/m2"),
        "enviscdiss": Varr("Viscous dissipation", "Heat flux", "W/m2"),
        "enadiabh": Varr("Adiabatic heating", "Heat flux", "W/m2"),
        "bsmean": Varr("Basalt content", "Concentration", "1"),
        "bsmin": Varr("Min basalt content", "Concentration", "1"),
        "bsmax": Varr("Max basalt content", "Concentration", "1"),
        "rhomean": Varr("Density", "Density", "kg/m3"),
        "rhomin": Varr("Min density", "Density", "kg/m3"),
        "rhomax": Varr("Max density", "Density", "kg/m3"),
        "airmean": Varr("Air", "Air", "1"),
        "airmin": Varr("Min air", "Air", "1"),
        "airmax": Varr("Max air", "Air", "1"),
        "primmean": Varr("Primordial", "Concentration", "1"),
        "primmin": Varr("Min primordial", "Concentration", "1"),
        "primmax": Varr("Max primordial", "Concentration", "1"),
        "ccmean": Varr("Continental crust", "Crust", "1"),
        "ccmin": Varr("Min continental crust", "Crust", "1"),
        "ccmax": Varr("Max continental crust", "Crust", "1"),
        "fmeltmean": Varr("Melt fraction", "Melt fraction", "1"),
        "fmeltmin": Varr("Min melt fraction", "Melt fraction", "1"),
        "fmeltmax": Varr("Max melt fraction", "Melt fraction", "1"),
        "metalmean": Varr("Metal", "Concentration", "1"),
        "metalmin": Varr("Min metal", "Concentration", "1"),
        "metalmax": Varr("Max metal", "Concentration", "1"),
        "gsmean": Varr("Grain size", "Grain size", "m"),
        "gsmin": Varr("Min grain size", "Grain size", "m"),
        "gsmax": Varr("Max grain", "Grain size", "m"),
        "viscdisslog": Varr("Viscous dissipation", "Heat flux", "W/m2"),
        "viscdissmin": Varr("Min visc dissipation", "Heat flux", "W/m2"),
        "viscdissmax": Varr("Max visc dissipation", "Heat flux", "W/m2"),
        "advtot": Varr("Advection", "Heat flux", "W/m2"),
        "advdesc": Varr("Downward advection", "Heat flux", "W/m2"),
        "advasc": Varr("Upward advection", "Heat flux", "W/m2"),
        "tcondmean": Varr("Conductivity", "Conductivity", "W/(m.K)"),
        "tcondmin": Varr("Min conductivity", "Conductivity", "W/(m.K)"),
        "tcondmax": Varr("Max conductivity", "Conductivity", "W/(m.K)"),
        "impmean": Varr("Impactor fraction", "Concentration", "1"),
        "impmin": Varr("Min impactor fraction", "Concentration", "1"),
        "impmax": Varr("Max impactor fraction", "Concentration", "1"),
        "hzmean": Varr("Harzburgite fraction", "Concentration", "1"),
        "hzmin": Varr("Min harzburgite fraction", "Concentration", "1"),
        "hzmax": Varr("Max harzburgite fraction", "Concentration", "1"),
        "TTGmean": Varr("TTG fraction", "Concentration", "1"),
        "TTGmin": Varr("Min TTG fraction", "Concentration", "1"),
        "TTGmax": Varr("Max TTG fraction", "Concentration", "1"),
        "edismean": Varr("Dislocation creep fraction", "Fraction", "1"),
        "edismin": Varr("Min dislocation creep fraction", "Fraction", "1"),
        "edismax": Varr("Max dislocation creep fraction", "Fraction", "1"),
        "egbsmean": Varr("Grain boundary sliding fraction", "Fraction", "1"),
        "egbsmin": Varr("Min grain boundary sliding fraction", "Fraction", "1"),
        "egbsmax": Varr("Max grain boundary sliding fraction", "Fraction", "1"),
        "ePeimean": Varr("Peierls creep fraction", "Fraction", "1"),
        "ePeimin": Varr("Min Peierls creep fraction", "Fraction", "1"),
        "ePeimax": Varr("Max Peierls creep fraction", "Fraction", "1"),
        "eplamean": Varr("Plasticity fraction", "Fraction", "1"),
        "eplamin": Varr("Min plasticity fraction", "Fraction", "1"),
        "eplamax": Varr("Max plasticity fraction", "Fraction", "1"),
    }
)
"""Radial profiles output by StagYY."""

RPROF_EXTRA: Mapping[str, Callable[[Step], Rprof]] = MappingProxyType(
    {
        "dr": processing.delta_r,
        "diff": processing.diff_prof,
        "diffs": processing.diffs_prof,
        "advts": processing.advts_prof,
        "advds": processing.advds_prof,
        "advas": processing.advas_prof,
        "energy": processing.energy_prof,
    }
)
"""Radial profiles that StagPy can compute."""


TIME: Mapping[str, Vart] = MappingProxyType(
    {
        "time": Vart("Time", "Time", "s"),
        "F_top": Vart("Heat flux at top", "Heat flux", "W/m2"),
        "F_bot": Vart("Heat flux at bottom", "Heat flux", "W/m2"),
        "Tmin": Vart("Min temperature", "Temperature", "K"),
        "Tmean": Vart("Temperature", "Temperature", "K"),
        "Tmax": Vart("Max temperature", "Temperature", "K"),
        "Vmin": Vart("Min velocity", "Velocity", "m/s"),
        "Vrms": Vart("rms velocity", "Velocity", "m/s"),
        "Vmax": Vart("Max velocity", "Velocity", "m/s"),
        "eta_min": Vart("Min viscosity", "Viscosity", "Pa.s"),
        "eta_amean": Vart("Viscosity", "Viscosity", "Pa.s"),
        "eta_max": Vart("Max viscosity", "Viscosity", "Pa.s"),
        "ra_eff": Vart("Effective Ra", r"$\mathrm{Ra}$", "1"),
        "Nu_top": Vart("Nusselt at top", r"$\mathrm{Nu}$", "1"),
        "Nu_bot": Vart("Nusselt at bot", r"$\mathrm{Nu}$", "1"),
        "C_min": Vart("Min concentration", "Concentration", "1"),
        "C_mean": Vart("Concentration", "Concentration", "1"),
        "C_max": Vart("Max concentration", "Concentration", "1"),
        "F_mean": Vart("Molten fraction", "Fraction", "1"),
        "F_max": Vart("Max molten fraction", "Fraction", "1"),
        "erupt_rate": Vart("Eruption rate", "Eruption rate", "1/s"),
        "erupta": Vart("Erupta total", "Eruption rate", "1/s"),
        "erupt_heatflux": Vart("Erupta heat", "Eruption rate", "1/s"),
        "entrainment": Vart("Entrainment", "Eruption rate", "1/s"),
        "Cmass_error": Vart("Error on Cmass", "Error", "1"),
        "H_int": Vart("Internal heating", "Power density", "W/m3"),
        "r_innercore": Vart("Inner core radius", "Inner core radius", "m"),
        "Tsurf": Vart("Temperature at top", "Temperature", "K"),
        "Tcmb": Vart("Temperature at bottom", "Temperature", "K"),
        "Tpotl": Vart("Mass mean potential temperature", "Temperature", "K"),
        "eta_gmean": Vart("Geometric mass mean viscosity", "Viscosity", "Pa.s"),
        "H_diffus": Vart("Heat diffusion", "Power density", "W/m3"),
        "H_VD": Vart("Viscous dissipation", "Power density", "W/m3"),
        "H_AH": Vart("Adiabatic heating", "Power density", "W/m3"),
        "H_cool": Vart("Cooling", "Heat flux", "W/m2"),
        "H_melt": Vart("Melting heat flux", "Heat flux", "W/m2"),
        "erupta_total": Vart("Total erupta", "Mass", "kg"),
        "Psurf": Vart("Ground pressure", "Pressure", "Pa"),
        "weathering": Vart("Weathering", "Mass flux", "tons/yr"),
        "mH2O_total": Vart("Total H2O mass", "Mass", "kg"),
        "mH2O_mantle": Vart("H2O in mantle", "Mass", "kg"),
        "mH2O_crust": Vart("H2O in crust", "Mass", "kg"),
        "dmH2O": Vart("Absorbed H2O", "Mass", "kg"),
        "cH2O_mean": Vart("H2O concentration", "Concentration", "1"),
        "F_top_H2O_diff": Vart("H2O diffusion flux at top", "Mass flux", "kg/s"),
        "F_top_H2O_erupted": Vart("H2O eruption flux at top", "Mass flux", "kg/s"),
        "outgassed_H2O": Vart("H2O outgassed", "Fraction", "1"),
        "cCO2_mean": Vart("CO2 concentration", "Concentration", "1"),
        "outgassed_CO2": Vart("CO2 outgassed", "Fraction", "1"),
        "cN2_mean": Vart("N2 concentration", "Concentration", "1"),
        "outgassed_N2": Vart("N2 outgassed", "Fraction", "1"),
        "c40Ar_mean": Vart("40Ar concentration", "Concentration", "1"),
        "c36Ar_mean": Vart("36Ar concentration", "Concentration", "1"),
        "outgassed_40Ar": Vart("40Ar outgassed", "Fraction", "1"),
        "outgassed_36Ar": Vart("36Ar outgassed", "Fraction", "1"),
        "c4He_mean": Vart("4He concentration", "Concentration", "1"),
        "c3He_mean": Vart("3He concentration", "Concentration", "1"),
        "outgassed_4He": Vart("4He outgassed", "Fraction", "1"),
        "outgassed_3He": Vart("3He outgassed", "Fraction", "1"),
        "IngassedH2O": Vart("Ingassed H2O", "Fraction", "1"),
        "OutputtedNotEruptedH2O": Vart("Outputted H2O", "Fraction", "1"),
        "SaturationOutgasH2O": Vart("Saturation-outgas H2O", "Fraction", "1"),
        "EruptedH2O": Vart("Erupted H2O", "Fraction", "1"),
        "SurfOceanMass3D": Vart("Scaled H2O mass", "Mass", "kg"),
        "TotH2OMassOceanPlusMantle": Vart("Total H2O mass", "Mass", "kg"),
        "GS_min": Vart("Min grain size", "Size", "1"),
        "GS_mean": Vart("Grain size", "Size", "1"),
        "GS_max": Vart("Max grain size", "Size", "1"),
        "intruda": Vart("Intruda", "Mantle mass", "1"),
        "erupta_TTG": Vart("TTG erupta", "Mantle mass", "1"),
        "intruda_TTG": Vart("TTG intruda", "Mantle mass", "1"),
        "TTG1": Vart("TTG1", "Mantle mass", "1"),
        "TTG2": Vart("TTG2", "Mantle mass", "1"),
        "TTG3": Vart("TTG3", "Mantle mass", "1"),
        "s_core": Vart("Core size", "Length", "m"),
        "tc_core": Vart("Temperature at center", "Temperature", "K"),
        "ts_core": Vart("Temperature at core surface", "Temperature", "K"),
        "m_Fe_S2TMO": Vart("FeO mass flux solid to TMO", "Mass", "kg"),
        "m_Fe_TMO2S": Vart("FeO mass flux TMO to solid", "Mass", "kg"),
        "m_Fe_S2BMO": Vart("FeO mass flux solid to BMO", "Mass", "kg"),
        "m_Fe_BMO2S": Vart("FeO mass flux BMO to solid", "Mass", "kg"),
        "c_Fe_TMO": Vart("FeO content TMO", "Concentration", "1"),
        "c_Fe_BMO": Vart("FeO content BMO", "Concentration", "1"),
        "c_Fe_sol": Vart("FeO content solid", "Concentration", "1"),
        "denstramin": Vart("Min tracers density", "Density", "kg/m3"),
        "denstramean": Vart("Tracers density", "Density", "kg/m3"),
        "denstramax": Vart("Max tracers density", "Density", "kg/m3"),
        "H_impacts": Vart("Impact heating", "Power", "W"),
        "SurfNetRotX": Vart("Surface net rotation around X", "Rotation", "1/s"),
        "SurfNetRotY": Vart("Surface net rotation around Y", "Rotation", "1/s"),
        "SurfNetRotZ": Vart("Surface net rotation around Z", "Rotation", "1/s"),
    }
)
"""Time series output by StagYY."""

TIME_EXTRA: Mapping[str, Callable[[StagyyData], Tseries]] = MappingProxyType(
    {
        "dt": processing.dtime,
        "dTdt": processing.dt_dt,
        "ebalance": processing.ebalance,
        "mobility": processing.mobility,
    }
)
"""Time series that StagPy can compute."""

TIME_ALIAS: Mapping[str, str] = MappingProxyType(
    {
        "t": "time",
        "ftop": "F_top",
        "fbot": "F_bot",
        "vmin": "Vmin",
        "vrms": "Vrms",
        "vmax": "Vmax",
        "etamin": "eta_min",
        "etamean": "eta_amean",
        "etamax": "eta_max",
        "Raeff": "ra_eff",
        "Nutop": "Nu_top",
        "Nubot": "Nu_bot",
        "Cmin": "C_min",
        "Cmean": "C_mean",
        "Cmax": "C_max",
        "moltenf_mean": "F_mean",
        "moltenf_max": "F_max",
        "erupt_tot": "erupta",
        "erupt_heat": "erupt_heatflux",
        "r_ic": "r_innercore",
        "topT_val": "Tsurf",
        "botT_val": "Tcmb",
    }
)
"""Time series name aliases."""

REFSTATE: Mapping[str, Varr] = MappingProxyType(
    {
        "z": Varr("z position", "z position", "m"),
        "T": Varr("Temperature", "Temperature", "K"),
        "rho": Varr("Density", "Density", "kg/m3"),
        "expan": Varr("Thermal expansivity", "Thermal expansivity", "1/K"),
        "Cp": Varr("Heat capacity", "Heat capacity", "J/(kg.K)"),
        "Tcond": Varr("Conductivity", "Conductivity", "W/(m.K)"),
        "P": Varr("Pressure", "Pressure", "Pa"),
        "grav": Varr("Gravity", "Gravity", "m/s2"),
    }
)
"""Variables in [`Refstate`][stagpy.stagyydata.Refstate]."""

SCALES: Mapping[str, Callable[[Scales], float]] = MappingProxyType(
    {
        "m": attrgetter("length"),
        "kg/m3": attrgetter("density"),
        "kg": attrgetter("mass"),
        "kg/s": lambda scl: scl.mass / scl.time,
        "K": attrgetter("temperature"),
        "W/m2": attrgetter("heat_flux"),
        "Pa": attrgetter("stress"),
        "Pa.s": attrgetter("dyn_visc"),
        "s": attrgetter("time"),
        "W": attrgetter("power"),
        "W/(m.K)": attrgetter("th_cond"),
        "m2/s": attrgetter("th_diff"),
        "W/m3": attrgetter("heat_production"),
        "1/s": lambda scl: 1 / scl.time,
        "K/s": lambda scl: scl.temperature / scl.time,
        "m/s": attrgetter("velocity"),
        "m/s2": attrgetter("acceleration"),
        "tons/yr": lambda scl: scl.mass / scl.time * 3.15e4,
    }
)
"""Scales to make values dimensional, see [`Scales`][stagpy.dimensions.Scales]."""

PREFIXES = ("k", "M", "G")



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from xml.etree import ElementTree as ET


class EndOfXml(Exception):
    """End of Xml has been reached."""


class XmlStream:
    def __init__(self, filepath: Path):
        self.filepath: Path = filepath
        self._event: str = "end"
        self._elem: ET.Element

    @cached_property
    def _cursor(self) -> Iterator[tuple[str, ET.Element]]:
        return ET.iterparse(str(self.filepath), events=("start", "end"))

    def _to_next_start(self) -> ET.Element:
        for self._event, self._elem in self._cursor:
            if self._event == "start":
                return self._elem
            self._elem.clear()
        raise EndOfXml("Reached end of file")

    @property
    def current(self) -> ET.Element:
        """Element at "start" event."""
        if self._event == "start":
            return self._elem
        return self._to_next_start()

    def advance(self) -> None:
        """Advance to next "start" event."""
        self.current  # make sure to be at current "start" event
        self._to_next_start()

    def skip_to_tag(self, tag: str) -> None:
        """Progress in file (both width and depth) until reaching the given tag."""
        while self.current.tag != tag:
            self.advance()

    def iter_tag(self, tag: str) -> Iterator[None]:
        try:
            while True:
                self.skip_to_tag(tag)
                yield None
        except EndOfXml:
            pass

    def iter_load_successive_tag(self, tag: str) -> Iterator[ET.Element]:
        try:
            while self.current.tag == tag:
                with self.load() as elt:
                    yield elt
        except EndOfXml:
            pass

    def drop(self) -> None:
        """Discard the current element and its children."""
        self.current  # make sure to be at current "start" event
        for self._event, elem in self._cursor:
            if self._event == "start":
                self.drop()
            else:
                elem.clear()
                return

    @contextmanager
    def load(self) -> Iterator[ET.Element]:
        """Fully read the current element and its children."""
        self.current  # make sure to be at current "start" event
        elem = None
        for self._event, elem in self._cursor:
            if self._event == "start":
                self.load().__enter__()
            else:
                yield elem
                break
        assert elem is not None
        elem.clear()



from __future__ import annotations

import typing
from dataclasses import dataclass
from functools import cached_property

import numpy as np

from . import phyvars
from .config import Scaling

if typing.TYPE_CHECKING:
    from typing import TypeVar

    from numpy.typing import NDArray

    from .parfile import StagyyPar
    from .stagyydata import StagyyData

    T = TypeVar("T", float, np.float64, NDArray[np.float64])


def apply_factors(data: T, unit: str, scaling: Scaling) -> tuple[T, str]:
    """Rescale a dimensional quantity according to scaling factors."""
    factor = scaling.factors.get(unit, "")
    if scaling.time_in_y and unit == "s":
        data /= scaling.yearins
        unit = "yr"
    elif scaling.vel_in_cmpy and unit == "m/s":
        data *= 100 * scaling.yearins
        unit = "cm/yr"
    if factor in phyvars.PREFIXES:
        data *= 10 ** (-3 * (phyvars.PREFIXES.index(factor) + 1))
        unit = factor + unit
    return data, unit


@dataclass(frozen=True)
class Scales:
    """Dimensional scales."""

    sdat: StagyyData

    @property
    def par(self) -> StagyyPar:
        return self.sdat.par

    def make_dimensional(self, data: T, unit: str, scaling: Scaling) -> tuple[T, str]:
        """Scale quantity to obtain dimensional quantity."""
        if self.par.get("switches", "dimensional_units", True) or unit == "1":
            return data, ""
        scale = phyvars.SCALES[unit](self)
        scale, unit = apply_factors(scale, unit, scaling)
        return data * scale, unit

    @cached_property
    def length(self) -> float:
        """Length in m."""
        thick = self.par.get("geometry", "d_dimensional", 2890e3)
        if self.par.get("boundaries", "air_layer", False):
            thick += self.par.nml["boundaries"]["air_thickness"]
        return thick

    @property
    def temperature(self) -> float:
        """Temperature in K."""
        return self.par.nml["refstate"]["deltaT_dimensional"]

    @property
    def density(self) -> float:
        """Density in kg/m3."""
        return self.par.nml["refstate"]["dens_dimensional"]

    @property
    def mass(self) -> float:
        """Mass in kg."""
        return self.density * self.length**3

    @property
    def th_cond(self) -> float:
        """Thermal conductivity in W/(m.K)."""
        return self.par.nml["refstate"]["tcond_dimensional"]

    @property
    def sp_heat(self) -> float:
        """Specific heat capacity in J/(kg.K)."""
        return self.par.nml["refstate"]["Cp_dimensional"]

    @property
    def dyn_visc(self) -> float:
        """Dynamic viscosity in Pa.s."""
        return self.par.nml["viscosity"]["eta0"]

    @property
    def th_diff(self) -> float:
        """Thermal diffusivity in m2/s."""
        return self.th_cond / (self.density * self.sp_heat)

    @property
    def time(self) -> float:
        """Time in s."""
        return self.length**2 / self.th_diff

    @property
    def velocity(self) -> float:
        """Velocity in m/s."""
        return self.length / self.time

    @property
    def acceleration(self) -> float:
        """Acceleration in m/s2."""
        return self.length / self.time**2

    @property
    def power(self) -> float:
        """Power in W."""
        return self.th_cond * self.temperature * self.length

    @property
    def heat_flux(self) -> float:
        """Local heat flux in W/m2."""
        return self.power / self.length**2

    @property
    def heat_production(self) -> float:
        """Local heat production in W/m3."""
        return self.power / self.length**3

    @property
    def stress(self) -> float:
        """Stress in Pa."""
        return self.dyn_visc / self.time



# pyright: reportDeprecated=false

"""Define configuration variables for StagPy.

See [stagpy.cli][] for additional definitions related to the command line
interface.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Sequence

import loam.parsers as lprs
from loam.base import ConfigBase, Section, entry
from loam.collections import MaybeEntry, TupleEntry
from loam.tools import command_flag, path_entry, switch_opt

_indices = TupleEntry(inner_from_toml=lprs.slice_or_int_parser)
_plots = TupleEntry.wrapping(
    TupleEntry.wrapping(TupleEntry(str), str_sep="."), str_sep="-"
)

CONFIG_LOCAL = Path(".stagpy.toml")
"""Path of local configuration file."""


@dataclass
class Common(Section):
    """General options."""

    config: bool = command_flag("print config options")


@dataclass
class Core(Section):
    """Core control."""

    path: Path = path_entry(
        path=".", cli_short="p", doc="path of StagYY run directory or par file"
    )
    read_parameters_dat: bool = switch_opt(True, None, "enable reading parameters.dat")
    outname: str = entry(val="stagpy", cli_short="n", doc="output file name prefix")
    shortname: bool = switch_opt(False, None, "output file name is only prefix")
    timesteps: Sequence[int | slice] = _indices.entry(
        doc="timesteps slice", in_file=False, cli_short="t"
    )
    snapshots: Sequence[int | slice] = _indices.entry(
        default=[-1], doc="snapshots slice", in_file=False, cli_short="s"
    )


@dataclass
class Plot(Section):
    """Plotting."""

    ratio: float | None = MaybeEntry(float).entry(
        doc="force aspect ratio of field plot", in_file=False
    )
    raster: bool = switch_opt(True, None, "rasterize field plots")
    format: str = entry(val="pdf", doc="figure format (pdf, eps, svg, png)")
    vmin: float | None = MaybeEntry(float).entry(
        doc="minimal value on plot", in_file=False
    )
    vmax: float | None = MaybeEntry(float).entry(
        doc="maximal value on plot", in_file=False
    )
    cminmax: bool = switch_opt(False, "C", "constant min max across plots")
    isolines: Sequence[float] = TupleEntry(float).entry(
        doc="list of isoline values", in_file=False
    )
    mplstyle: Sequence[str] = TupleEntry(str).entry(
        default="stagpy-paper", doc="list of matplotlib styles", in_file=False
    )
    xkcd: bool = command_flag("use the xkcd style")


@dataclass
class Scaling(Section):
    """Dimensionalization."""

    yearins: float = entry(val=3.154e7, in_cli=False, doc="year in seconds")
    ttransit: float = entry(val=1.78e15, in_cli=False, doc="transit time in My")
    time_in_y: bool = switch_opt(True, None, "dimensional time is in year")
    vel_in_cmpy: bool = switch_opt(True, None, "dimensional velocity is in cm/year")
    factors: Dict[str, str] = entry(
        val_factory=lambda: {"s": "M", "m": "k", "Pa": "G"},
        in_cli=False,
        doc="custom factors",
    )


@dataclass
class FieldSec(Section):
    """Field command."""

    plot: Sequence[Sequence[Sequence[str]]] = _plots.entry(
        default="T,stream", cli_short="o", doc="variables to plot (see stagpy var)"
    )
    perturbation: bool = switch_opt(False, None, "plot departure from average profile")
    shift: int | None = MaybeEntry(int).entry(
        doc="shift plot horizontally", in_file=False
    )
    timelabel: bool = switch_opt(False, None, "add label with time")
    colorbar: bool = switch_opt(True, None, "add color bar to plot")
    ix: int | None = MaybeEntry(int).entry(
        doc="x-index of slice for 3D fields", in_file=False
    )
    iy: int | None = MaybeEntry(int).entry(
        doc="y-index of slice for 3D fields", in_file=False
    )
    iz: int | None = MaybeEntry(int).entry(
        doc="z-index of slice for 3D fields", in_file=False
    )
    isocolors: Sequence[str] = TupleEntry(str).entry(doc="list of colors for isolines")
    cmap: Dict[str, str] = entry(
        val_factory=lambda: {
            "T": "RdBu_r",
            "eta": "viridis_r",
            "rho": "RdBu",
            "sII": "plasma_r",
            "edot": "Reds",
        },
        in_cli=False,
        doc="custom colormaps",
    )


@dataclass
class RprofSec(Section):
    """Rprof command."""

    plot: Sequence[Sequence[Sequence[str]]] = _plots.entry(
        default="Tmean", cli_short="o", doc="variables to plot (see stagpy var)"
    )
    style: str = entry(val="-", doc="matplotlib line style")
    average: bool = switch_opt(False, "a", "plot temporal average")
    grid: bool = switch_opt(False, "g", "plot grid")
    depth: bool = switch_opt(False, "d", "depth as vertical axis")


@dataclass
class Time(Section):
    """Time command."""

    plot: Sequence[Sequence[Sequence[str]]] = _plots.entry(
        default="Nu_top,Nu_bot.Vrms.Tmean",
        cli_short="o",
        doc="variables to plot (see stagpy var)",
    )
    style: str = entry(val="-", doc="matplotlib line style")
    compstat: Sequence[str] = TupleEntry(str).entry(
        doc="compute mean and rms of listed variables", in_file=False
    )
    tstart: float | None = MaybeEntry(float).entry(doc="beginning time", in_file=False)
    tend: float | None = MaybeEntry(float).entry(doc="end time", in_file=False)
    fraction: float | None = MaybeEntry(float).entry(
        doc="ending fraction of series to process", in_file=False
    )
    marktimes: Sequence[float] = TupleEntry(float).entry(
        doc="list of times where to put a mark", in_file=False, cli_short="M"
    )
    marksteps: Sequence[int | slice] = _indices.entry(
        doc="list of steps where to put a mark", in_file=False, cli_short="T"
    )
    marksnaps: Sequence[int | slice] = _indices.entry(
        doc="list of snaps where to put a mark", in_file=False, cli_short="S"
    )


@dataclass
class Refstate(Section):
    """Refstate command."""

    plot: Sequence[str] = TupleEntry(str).entry(
        default="T", cli_short="o", doc="variables to plot (see stagpy var)"
    )
    style: str = entry(val="-", doc="matplotlib line style")


@dataclass
class Plates(Section):
    """Plates command."""

    plot: Sequence[Sequence[Sequence[str]]] = _plots.entry(
        default="c.T.v2-v2.dv2-v2.topo_top",
        cli_short="o",
        doc="variables to plot, can be a surface field, field, or dv2",
    )
    field: str = entry(val="eta", doc="field to plot with plates info")
    stress: bool = switch_opt(
        False, None, "plot deviatoric stress instead of velocity on field plots"
    )
    continents: bool = switch_opt(True, None, "whether to shade continents")
    vzratio: float = entry(
        val=0.0, doc="Ratio of mean vzabs used as threshold for plates limits"
    )
    nbplates: bool = switch_opt(
        False, None, "plot number of plates as function of time"
    )
    distribution: bool = switch_opt(False, None, "plot plate size distribution")
    zoom: float | None = MaybeEntry(float).entry(
        doc="zoom around surface", in_file=False
    )


@dataclass
class Info(Section):
    """Info command."""

    output: Sequence[str] = TupleEntry(str).entry(
        default="time,Tmean,Vrms,Nu_top,Nu_bot",
        cli_short="o",
        doc="time series to print",
    )


@dataclass
class Var(Section):
    """Var command."""

    field: bool = command_flag("print field variables")
    sfield: bool = command_flag("print surface field variables")
    rprof: bool = command_flag("print rprof variables")
    time: bool = command_flag("print time variables")
    refstate: bool = command_flag("print refstate variables")


@dataclass
class ConfSection(Section):
    """Config command."""

    create: bool = command_flag("create config file")


@dataclass
class Completions(Section):
    """Shell completion scripts."""

    bash: bool = command_flag("generate bash completion")
    zsh: bool = command_flag("generate zsh completion")


@dataclass
class Config(ConfigBase):
    """StagPy configuration."""

    common: Common
    core: Core
    plot: Plot
    scaling: Scaling
    field: FieldSec
    rprof: RprofSec
    time: Time
    refstate: Refstate
    plates: Plates
    info: Info
    var: Var
    config: ConfSection
    completions: Completions



"""Plot scalar and vector fields."""

from __future__ import annotations

import typing
from itertools import chain

import matplotlib.colors as mpl_colors
import matplotlib.patches as mpat
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from stagpy.dimensions import apply_factors

from . import _helpers, phyvars
from .config import Config
from .error import NotAvailableError
from .stagyydata import _sdat_from_conf

if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any

    from matplotlib.axes import Axes
    from matplotlib.collections import QuadMesh
    from matplotlib.colorbar import Colorbar
    from matplotlib.figure import Figure
    from numpy.typing import NDArray

    from .datatypes import Varf
    from .stagyydata import StepsView
    from .step import Step


# The location is off for vertical velocities: they have an extra
# point in (x,y) instead of z in the output


def _threed_extract(
    conf: Config, step: Step, var: str, walls: bool = False
) -> tuple[tuple[NDArray[np.float64], NDArray[np.float64]], NDArray[np.float64]]:
    """Return suitable slices and coords for 3D fields."""
    is_vector = not valid_field_var(var)
    hwalls = is_vector or walls
    i_x: int | slice | None = conf.field.ix
    i_y: int | slice | None = conf.field.iy
    i_z: int | slice | None = conf.field.iz
    if i_x is not None or i_y is not None:
        i_z = None
    if i_x is not None or i_z is not None:
        i_y = None
    if i_x is None and i_y is None and i_z is None:
        i_x = 0
    if i_x is not None:
        xcoord = step.geom.y_walls if hwalls else step.geom.y_centers
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
        i_y = i_z = slice(None)
        varx, vary = var + "2", var + "3"
    elif i_y is not None:
        xcoord = step.geom.x_walls if hwalls else step.geom.x_centers
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
        i_x = i_z = slice(None)
        varx, vary = var + "1", var + "3"
    else:
        xcoord = step.geom.x_walls if hwalls else step.geom.x_centers
        ycoord = step.geom.y_walls if hwalls else step.geom.y_centers
        i_x = i_y = slice(None)
        varx, vary = var + "1", var + "2"
    data: Any
    if is_vector:
        data = (
            step.fields[varx].values[i_x, i_y, i_z, 0],
            step.fields[vary].values[i_x, i_y, i_z, 0],
        )
    else:
        data = step.fields[var].values[i_x, i_y, i_z, 0]
    return (xcoord, ycoord), data


def valid_field_var(var: str) -> bool:
    """Whether a field variable is defined.

    Args:
        var: the variable name to be checked.

    Returns:
        whether the var is defined in either [`FIELD`][stagpy.phyvars.FIELD] or
            [`FIELD_EXTRA`][stagpy.phyvars.FIELD_EXTRA].
    """
    return var in phyvars.FIELD or var in phyvars.FIELD_EXTRA


def get_meshes_fld(
    conf: Config, step: Step, var: str, walls: bool = False
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], Varf]:
    """Return scalar field along with coordinates meshes.

    Only works properly in 2D geometry and 3D cartesian.

    Args:
        conf: configuration.
        step: a `Step` of a `StagyyData` instance.
        var: scalar field name.
        walls: consider the walls as the relevant mesh.

    Returns:
        xmesh: x position
        ymesh: y position
        fld: field values
        meta: metadata
    """
    fld = step.fields[var]
    hwalls = (
        walls
        or fld.values.shape[0] != step.geom.nxtot
        or fld.values.shape[1] != step.geom.nytot
    )

    if step.geom.curvilinear:
        if step.geom.twod_yz:
            xcoord = step.geom.y_walls if hwalls else step.geom.y_centers
            ycoord = step.geom.z_walls if walls else step.geom.z_centers
            pmesh, rmesh = np.meshgrid(xcoord, ycoord, indexing="ij")
            xmesh, ymesh = rmesh * np.cos(pmesh), rmesh * np.sin(pmesh)
            vals = fld.values[0, :, :, 0]
            return xmesh, ymesh, vals, fld.meta
        else:
            raise NotImplementedError()

    # cartesian
    if step.geom.threed:
        (xcoord, ycoord), vals = _threed_extract(conf, step, var, walls)
    else:
        if step.geom.twod_xz:
            xcoord = step.geom.x_walls if hwalls else step.geom.x_centers
            vals = fld.values[:, 0, :, 0]
        else:  # twod_yz
            xcoord = step.geom.y_walls if hwalls else step.geom.y_centers
            vals = fld.values[0, :, :, 0]
        ycoord = step.geom.z_walls if walls else step.geom.z_centers
    xmesh, ymesh = np.meshgrid(xcoord, ycoord, indexing="ij")
    return xmesh, ymesh, vals, fld.meta


def get_meshes_vec(
    conf: Config, step: Step, var: str
) -> tuple[
    NDArray[np.float64], NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]
]:
    """Return vector field components along with coordinates meshes.

    Only works properly in 2D geometry and 3D cartesian.

    Args:
        conf: configuration.
        step: a `Step` of a `StagyyData` instance.
        var: vector field name.

    Returns:
        xmesh: x position
        ymesh: y position
        fldx: x component
        fldy: y component
    """
    if step.geom.curvilinear:
        if step.geom.twod_yz:
            pcoord = step.geom.p_walls
            pmesh = np.outer(pcoord, np.ones(step.geom.nrtot))
            vec_phi = step.fields[var + "2"].values[0, :, :, 0]
            vec_r = step.fields[var + "3"].values[0, :, :, 0]
            vec1 = vec_r * np.cos(pmesh) - vec_phi * np.sin(pmesh)
            vec2 = vec_phi * np.cos(pmesh) + vec_r * np.sin(pmesh)
            pcoord, rcoord = step.geom.p_walls, step.geom.r_centers
            pmesh, rmesh = np.meshgrid(pcoord, rcoord, indexing="ij")
            xmesh, ymesh = rmesh * np.cos(pmesh), rmesh * np.sin(pmesh)
            return xmesh, ymesh, vec1, vec2
        else:
            raise NotImplementedError()

    # cartesian
    if step.geom.threed:
        (xcoord, ycoord), (vec1, vec2) = _threed_extract(conf, step, var)
    elif step.geom.twod_xz:
        xcoord, ycoord = step.geom.x_walls, step.geom.z_centers
        vec1 = step.fields[var + "1"].values[:, 0, :, 0]
        vec2 = step.fields[var + "3"].values[:, 0, :, 0]
    else:  # twod_yz
        xcoord, ycoord = step.geom.y_walls, step.geom.z_centers
        vec1 = step.fields[var + "2"].values[0, :, :, 0]
        vec2 = step.fields[var + "3"].values[0, :, :, 0]
    xmesh, ymesh = np.meshgrid(xcoord, ycoord, indexing="ij")
    return xmesh, ymesh, vec1, vec2


def plot_scalar(
    step: Step,
    var: str,
    field: NDArray[np.float64] | None = None,
    axis: Axes | None = None,
    conf: Config | None = None,
    **extra: Any,
) -> tuple[Figure, Axes, QuadMesh, Colorbar | None]:
    """Plot scalar field.

    Args:
        step: a `Step` of a `StagyyData` instance.
        var: the scalar field name.
        field: if not None, it is plotted instead of `step.fields[var]`.  This is
            useful to plot a masked or rescaled array.
        axis: the `matplotlib.axes.Axes` object where the field should
            be plotted.  If set to None, a new figure with one subplot is
            created.
        conf: configuration.
        extra: options that will be passed on to `matplotlib.axes.Axes.pcolormesh`.

    Returns:
        fig: matplotlib figure
        axes: matplotlib axes
        surf: surface returned by `pcolormesh`
        cbar: colorbar
    """
    if conf is None:
        conf = Config.default_()
    if step.geom.threed and step.geom.spherical:
        raise NotAvailableError("plot_scalar not implemented for 3D spherical geometry")

    xmesh, ymesh, fld, meta = get_meshes_fld(conf, step, var, walls=True)
    # interpolate at cell centers, this should be abstracted by field objects
    # via an "at_cell_centers" method or similar
    if fld.shape[0] > max(step.geom.nxtot, step.geom.nytot):
        fld: NDArray[np.float64] = (fld[:-1] + fld[1:]) / 2  # type: ignore

    xmin, xmax = xmesh.min(), xmesh.max()
    ymin, ymax = ymesh.min(), ymesh.max()

    if field is not None:
        fld = field
    if conf.field.perturbation:
        fld = fld - np.mean(fld, axis=0)
    if conf.field.shift:
        fld = np.roll(fld, conf.field.shift, axis=0)

    if axis is None:
        fig, axis = plt.subplots(ncols=1)
    else:
        fig = axis.get_figure()  # type: ignore

    if step.sdat.par.get("magma_oceans_in", "evolving_magma_oceans", False):
        rcmb = step.sdat.par.nml["geometry"]["r_cmb"]
        xmax = rcmb + 1
        ymax = xmax
        xmin = -xmax
        ymin = -ymax
        rsurf = xmax if step.timeinfo["thick_tmo"] > 0 else step.geom.r_walls[-3]
        cmb = mpat.Circle((0, 0), rcmb, color="dimgray", zorder=0)
        psurf = mpat.Circle((0, 0), rsurf, color="indianred", zorder=0)
        axis.add_patch(psurf)
        axis.add_patch(cmb)

    extra_opts = dict(
        cmap=conf.field.cmap.get(var),
        vmin=conf.plot.vmin,
        vmax=conf.plot.vmax,
        norm=mpl_colors.LogNorm() if var == "eta" else None,
        rasterized=conf.plot.raster,
        shading="flat",
    )
    extra_opts.update(extra)
    surf = axis.pcolormesh(xmesh, ymesh, fld, **extra_opts)  # type: ignore

    cbar = None
    if conf.field.colorbar:
        cax = make_axes_locatable(axis).append_axes("right", size="3%", pad=0.15)
        cbar = plt.colorbar(surf, cax=cax)
        cbar.set_label(meta.description + (" pert." if conf.field.perturbation else ""))
    if step.geom.spherical or conf.plot.ratio is None:
        axis.set_aspect("equal")
        axis.set_axis_off()
    else:
        axis.set_aspect(conf.plot.ratio / axis.get_data_ratio())

    axis.set_adjustable("box")
    axis.set_xlim(xmin, xmax)
    axis.set_ylim(ymin, ymax)
    return fig, axis, surf, cbar


def plot_iso(
    axis: Axes,
    step: Step,
    var: str,
    field: NDArray[np.float64] | None = None,
    conf: Config | None = None,
    **extra: Any,
) -> None:
    """Plot isocontours of scalar field.

    Args:
        axis: the `matplotlib.axes.Axes` of an existing matplotlib
            figure where the isocontours should be plotted.
        step: a `Step` of a `StagyyData` instance.
        var: the scalar field name.
        field: if not None, it is plotted instead of `step.fields[var]`.  This is
            useful to plot a masked or rescaled array.
        conf: configuration.
        extra: options that will be passed on to `Axes.contour`.
    """
    if conf is None:
        conf = Config.default_()
    xmesh, ymesh, fld, _ = get_meshes_fld(conf, step, var)

    if field is not None:
        fld = field

    if conf.field.shift:
        fld = np.roll(fld, conf.field.shift, axis=0)
    extra_opts: dict[str, Any] = dict(linewidths=1)
    if "cmap" not in extra and conf.field.isocolors:
        extra_opts["colors"] = conf.field.isocolors
    elif "colors" not in extra:
        extra_opts["cmap"] = conf.field.cmap.get(var)
    if conf.plot.isolines:
        extra_opts["levels"] = sorted(conf.plot.isolines)
    extra_opts.update(extra)
    axis.contour(xmesh, ymesh, fld, **extra_opts)


def plot_vec(
    axis: Axes,
    step: Step,
    var: str,
    conf: Config | None = None,
) -> None:
    """Plot vector field.

    Args:
        axis: the :class:`matplotlib.axes.Axes` of an existing matplotlib
            figure where the vector field should be plotted.
        step: a `Step` of a `StagyyData` instance.
        var: the vector field name.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    xmesh, ymesh, vec1, vec2 = get_meshes_vec(conf, step, var)
    dipz = step.geom.nztot // 10
    if conf.field.shift:
        vec1 = np.roll(vec1, conf.field.shift, axis=0)
        vec2 = np.roll(vec2, conf.field.shift, axis=0)
    if step.geom.spherical or conf.plot.ratio is None:
        dipx = dipz
    else:
        dipx = step.geom.nytot if step.geom.twod_yz else step.geom.nxtot
        dipx = int(dipx // 10 * conf.plot.ratio) + 1
    axis.quiver(
        xmesh[::dipx, ::dipz],
        ymesh[::dipx, ::dipz],
        vec1[::dipx, ::dipz],
        vec2[::dipx, ::dipz],
        linewidths=1,
    )


def _findminmax(view: StepsView, sovs: Iterable[str]) -> dict[str, tuple[float, float]]:
    """Find min and max values of several fields."""
    minmax: dict[str, tuple[float, float]] = {}
    for step in view.filter(snap=True):
        for var in sovs:
            if var in step.fields:
                vals = step.fields[var].values
                if var in minmax:
                    minmax[var] = (
                        min(minmax[var][0], np.nanmin(vals).item()),
                        max(minmax[var][1], np.nanmax(vals).item()),
                    )
                else:
                    minmax[var] = np.nanmin(vals).item(), np.nanmax(vals).item()
    return minmax


def cmd(conf: Config) -> None:
    """Plot scalar and vector fields.

    This is the implementation of the `field` subcommand.
    """
    sdat = _sdat_from_conf(conf.core)
    view = _helpers.walk(sdat, conf)
    # no more than two fields in a subplot
    lovs = [[slov[:2] for slov in plov] for plov in conf.field.plot]
    minmax = {}
    if conf.plot.cminmax:
        conf.plot.vmin = None
        conf.plot.vmax = None
        sovs = set(slov[0] for plov in lovs for slov in plov)
        minmax = _findminmax(view, sovs)
    for step in view.filter(snap=True):
        for vfig in lovs:
            fig, axes = plt.subplots(
                ncols=len(vfig), squeeze=False, figsize=(6 * len(vfig), 6)
            )
            for axis, var in zip(axes[0], vfig):
                if var[0] not in step.fields:
                    print(f"{var[0]!r} field on snap {step.isnap} not found")
                    continue
                opts: dict[str, Any] = {}
                if var[0] in minmax:
                    opts = dict(vmin=minmax[var[0]][0], vmax=minmax[var[0]][1])
                plot_scalar(step, var[0], axis=axis, conf=conf, **opts)
                if len(var) == 2:
                    if valid_field_var(var[1]):
                        plot_iso(axis, step, var[1], conf=conf)
                    elif valid_field_var(var[1] + "1"):
                        plot_vec(axis, step, var[1], conf=conf)
            if conf.field.timelabel:
                time = step.time
                unit = ""
                if sdat.par.get("switches", "dimensional_units", True):
                    time, unit = apply_factors(time, "s", conf.scaling)
                    unit = " " + unit
                time_str = _helpers.scilabel(time)
                axes[0, 0].text(
                    0.02, 1.02, f"$t={time_str}${unit}", transform=axes[0, 0].transAxes
                )
            oname = "_".join(chain.from_iterable(vfig))
            plt.tight_layout(w_pad=3)
            _helpers.saveplot(conf, fig, oname, step.isnap)
